@@ -65,7 +65,7 @@
 #include "winmenu.h"
 #include "osdep.h"
 
-#ifdef MWM_HINTS
+#ifdef USE_MWM_HINTS
 # include "motif.h"
 #endif
 #include "wmspec.h"
@@ -178,14 +178,14 @@ void wWindowDestroy(WWindow *wwin)
 	wwin->flags.destroyed = 1;
 
 	for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
-		if (!w_global.shortcut.windows[i])
+		if (!wwin->screen_ptr->shortcutWindows[i])
 			continue;
 
-		WMRemoveFromArray(w_global.shortcut.windows[i], wwin);
+		WMRemoveFromArray(wwin->screen_ptr->shortcutWindows[i], wwin);
 
-		if (!WMGetArrayItemCount(w_global.shortcut.windows[i])) {
-			WMFreeArray(w_global.shortcut.windows[i]);
-			w_global.shortcut.windows[i] = NULL;
+		if (!WMGetArrayItemCount(wwin->screen_ptr->shortcutWindows[i])) {
+			WMFreeArray(wwin->screen_ptr->shortcutWindows[i]);
+			wwin->screen_ptr->shortcutWindows[i] = NULL;
 		}
 	}
 
@@ -288,12 +288,39 @@ static void setupGNUstepHints(WWindow *wwin, GNUstepWMAttributes *gs_hints)
 		wwin->client_flags.no_appicon = 1;
 }
 
+static void discard_hints_from_gtk(WWindow *wwin)
+{
+	Atom type;
+	int format;
+	unsigned long nb_item, nb_remain;
+	unsigned char *result;
+	int status;
+
+	status = XGetWindowProperty(dpy, wwin->client_win, w_global.atom.desktop.gtk_object_path, 0, 16, False,
+	                            AnyPropertyType, &type, &format, &nb_item, &nb_remain, &result);
+	if (status != Success)
+		return;
+
+	/* If we're here, that means the Property exists. We don't care what is inside, it means it is a GTK-based application */
+
+	if (result != NULL)
+		XFree(result);
+
+	/* GTK is asking to remove these decorations: */
+	wwin->client_flags.no_titlebar = 0;
+	wwin->client_flags.no_close_button = 0;
+	wwin->client_flags.no_miniaturize_button = 0;
+	wwin->client_flags.no_resizebar = 0;
+}
+
 void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 {
 	WScreen *scr = wwin->screen_ptr;
 
 	/* sets global default stuff */
-	wDefaultFillAttributes(wwin->wm_instance, wwin->wm_class, &wwin->client_flags, NULL, True);
+	wDefaultFillAttributes(wwin->wm_instance, wwin->wm_class, &wwin->user_flags, NULL, True);
+	wwin->defined_user_flags = wwin->user_flags;
+
 	/*
 	 * Decoration setting is done in this precedence (lower to higher)
 	 * - use global default in the resource database
@@ -302,25 +329,25 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 	 * - set hints specified for the app in the resource DB
 	 *
 	 */
-	WSETUFLAG(wwin, broken_close, 0);
+	wwin->client_flags.broken_close = 0;
 
 	if (wwin->protocols.DELETE_WINDOW)
-		WSETUFLAG(wwin, kill_close, 0);
+		wwin->client_flags.kill_close = 0;
 	else
-		WSETUFLAG(wwin, kill_close, 1);
+		wwin->client_flags.kill_close = 1;
 
 	/* transients can't be iconified or maximized */
 	if (wwin->transient_for != None && wwin->transient_for != scr->root_win) {
-		WSETUFLAG(wwin, no_miniaturizable, 1);
-		WSETUFLAG(wwin, no_miniaturize_button, 1);
+		wwin->client_flags.no_miniaturizable = 1;
+		wwin->client_flags.no_miniaturize_button = 1;
 	}
 
 	/* if the window can't be resized, remove the resizebar */
 	if (wwin->normal_hints->flags & (PMinSize | PMaxSize)
 	    && (wwin->normal_hints->min_width == wwin->normal_hints->max_width)
 	    && (wwin->normal_hints->min_height == wwin->normal_hints->max_height)) {
-		WSETUFLAG(wwin, no_resizable, 1);
-		WSETUFLAG(wwin, no_resizebar, 1);
+		wwin->client_flags.no_resizable = 1;
+		wwin->client_flags.no_resizebar = 1;
 	}
 
 	/* set GNUstep window attributes */
@@ -343,12 +370,15 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 		int tmp_workspace = -1;
 		int tmp_level = INT_MIN;	/* INT_MIN is never used by the window levels */
 
-#ifdef MWM_HINTS
+#ifdef USE_MWM_HINTS
 		wMWMCheckClientHints(wwin);
-#endif				/* MWM_HINTS */
+#endif	/* USE_MWM_HINTS */
 
 		wNETWMCheckClientHints(wwin, &tmp_level, &tmp_workspace);
 
+		if (wPreferences.ignore_gtk_decoration_hints)
+			discard_hints_from_gtk(wwin);
+
 		/* window levels are between INT_MIN+1 and INT_MAX, so if we still
 		 * have INT_MIN that means that no window level was requested. -Dan
 		 */
@@ -373,7 +403,7 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 		}
 
 		if (tmp_workspace >= 0)
-			*workspace = tmp_workspace % w_global.workspace.count;
+			*workspace = tmp_workspace % scr->workspace_count;
 	}
 
 	/*
@@ -401,7 +431,9 @@ void wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 	    && wwin->user_flags.floating && wwin->defined_user_flags.floating)
 		wwin->user_flags.sunken = 0;
 
-	WSETUFLAG(wwin, no_shadeable, WFLAGP(wwin, no_titlebar));
+	/* A window that does not have a title cannot be Shaded and we don't let user override this */
+	wwin->client_flags.no_shadeable = WFLAGP(wwin, no_titlebar);
+	wwin->defined_user_flags.no_shadeable = 0;
 
 	/* windows that have takefocus=False shouldn't take focus at all */
 	if (wwin->focus_mode == WFM_NO_INPUT)
@@ -724,22 +756,22 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	wwin->orig_main_window = wwin->main_window;
 
 	if (wwin->flags.is_gnustep)
-		WSETUFLAG(wwin, shared_appicon, 0);
+		wwin->client_flags.shared_appicon = 0;
 
 	if (wwin->main_window) {
 		XTextProperty text_prop;
 
 		if (XGetTextProperty(dpy, wwin->main_window, &text_prop, w_global.atom.wmaker.menu))
-			WSETUFLAG(wwin, shared_appicon, 0);
+			wwin->client_flags.shared_appicon = 0;
 	}
 
 	if (wwin->flags.is_dockapp)
-		WSETUFLAG(wwin, shared_appicon, 0);
+		wwin->client_flags.shared_appicon = 0;
 
 	if (wwin->main_window) {
             WApplication *app = wApplicationOf(wwin->main_window);
             if (app && app->app_icon)
-		WSETUFLAG(wwin, shared_appicon, 0);
+		wwin->client_flags.shared_appicon = 0;
         }
 
 	if (!withdraw && wwin->main_window && WFLAGP(wwin, shared_appicon)) {
@@ -749,7 +781,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 #define ADEQUATE(x) ((x)!=None && (x)!=wwin->client_win && (x)!=fPtr->leader)
 
-		/* // only enter here if PropGetWMClass() succeds */
+		/* // only enter here if PropGetWMClass() succeeds */
 		PropGetWMClass(wwin->main_window, &class, &instance);
 		buffer = StrConcatDot(instance, class);
 
@@ -833,16 +865,17 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 			wwin->flags.miniaturized = win_state->state->miniaturized;
 
 		if (!IS_OMNIPRESENT(wwin)) {
-			int w = wDefaultGetStartWorkspace(wwin->wm_instance, wwin->wm_class);
-			if (w < 0 || w >= w_global.workspace.count) {
+			int w = wDefaultGetStartWorkspace(scr, wwin->wm_instance,
+							  wwin->wm_class);
+			if (w < 0 || w >= scr->workspace_count) {
 				workspace = win_state->state->workspace;
-				if (workspace >= w_global.workspace.count)
-					workspace = w_global.workspace.current;
+				if (workspace >= scr->workspace_count)
+					workspace = scr->current_workspace;
 			} else {
 				workspace = w;
 			}
 		} else {
-			workspace = w_global.workspace.current;
+			workspace = scr->current_workspace;
 		}
 	}
 
@@ -883,10 +916,10 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 				for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
 					if (mask & (1 << i)) {
-						if (!w_global.shortcut.windows[i])
-							w_global.shortcut.windows[i] = WMCreateArray(4);
+						if (!scr->shortcutWindows[i])
+							scr->shortcutWindows[i] = WMCreateArray(4);
 
-						WMAddToArray(w_global.shortcut.windows[i], wwin);
+						WMAddToArray(scr->shortcutWindows[i], wwin);
 					}
 				}
 			}
@@ -905,20 +938,20 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 
 	/* set workspace on which the window starts */
 	if (workspace >= 0) {
-		if (workspace > w_global.workspace.count - 1)
-			workspace = workspace % w_global.workspace.count;
+		if (workspace > scr->workspace_count - 1)
+			workspace = workspace % scr->workspace_count;
 	} else {
 		int w;
 
-		w = wDefaultGetStartWorkspace(wwin->wm_instance, wwin->wm_class);
+		w = wDefaultGetStartWorkspace(scr, wwin->wm_instance, wwin->wm_class);
 
-		if (w >= 0 && w < w_global.workspace.count && !(IS_OMNIPRESENT(wwin))) {
+		if (w >= 0 && w < scr->workspace_count && !(IS_OMNIPRESENT(wwin))) {
 			workspace = w;
 		} else {
 			if (wPreferences.open_transients_with_parent && transientOwner)
 				workspace = transientOwner->frame->workspace;
 			else
-				workspace = w_global.workspace.current;
+				workspace = scr->current_workspace;
 		}
 	}
 
@@ -931,9 +964,8 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	wWindowConstrainSize(wwin, &width, &height);
 
 	/* do not ask for window placement if the window is
-	 * transient, during startup, if the initial workspace is another one
-	 * or if the window wants to start iconic.
-	 * If geometry was saved, restore it. */
+	 * transient, during startup, or if the window wants
+	 * to start iconic.  If geometry was saved, restore it. */
 	{
 		Bool dontBring = False;
 
@@ -942,7 +974,6 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 			y = win_state->state->y;
 		} else if ((wwin->transient_for == None || wPreferences.window_placement != WPM_MANUAL)
 			   && !scr->flags.startup
-			   && workspace == w_global.workspace.current
 			   && !wwin->flags.miniaturized
 			   && !wwin->flags.maximized && !(wwin->normal_hints->flags & (USPosition | PPosition))) {
 
@@ -1183,7 +1214,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	XLowerWindow(dpy, window);
 
 	/* if window is in this workspace and should be mapped, then  map it */
-	if (!wwin->flags.miniaturized && (workspace == w_global.workspace.current || IS_OMNIPRESENT(wwin))
+	if (!wwin->flags.miniaturized && (workspace == scr->current_workspace || IS_OMNIPRESENT(wwin))
 	    && !wwin->flags.hidden && !withdraw) {
 
 		/* The following "if" is to avoid crashing of clients that expect
@@ -1208,7 +1239,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 		     * the window focusing (if the mouse is over that window)
 		     * will be processed by wmaker.
 		     * But since this event will be rather delayed
-		     * (step 3 has a large delay) the time when the event ocurred
+		     * (step 3 has a large delay) the time when the event occurred
 		     * and when it is processed, the client that owns that window
 		     * will reject the XSetInputFocus() for it.
 		     */
@@ -1257,7 +1288,7 @@ WWindow *wManageWindow(WScreen *scr, Window window)
 	/* Final preparations before window is ready to go */
 	wFrameWindowChangeState(wwin->frame, WS_UNFOCUSED);
 
-	if (!wwin->flags.miniaturized && workspace == w_global.workspace.current && !wwin->flags.hidden) {
+	if (!wwin->flags.miniaturized && workspace == scr->current_workspace && !wwin->flags.hidden) {
 		if (((transientOwner && transientOwner->flags.focused)
 		     || wPreferences.auto_focus) && !WFLAGP(wwin, no_focusable)) {
 
@@ -1319,10 +1350,10 @@ WWindow *wManageInternalWindow(WScreen *scr, Window window, Window owner,
 
 	wwin->flags.internal_window = 1;
 
-	WSETUFLAG(wwin, omnipresent, 1);
-	WSETUFLAG(wwin, no_shadeable, 1);
-	WSETUFLAG(wwin, no_resizable, 1);
-	WSETUFLAG(wwin, no_miniaturizable, 1);
+	wwin->client_flags.omnipresent = 1;
+	wwin->client_flags.no_shadeable = 1;
+	wwin->client_flags.no_resizable = 1;
+	wwin->client_flags.no_miniaturizable = 1;
 
 	wwin->focus_mode = WFM_PASSIVE;
 	wwin->client_win = window;
@@ -1366,7 +1397,7 @@ WWindow *wManageInternalWindow(WScreen *scr, Window window, Window owner,
 	wFrameWindowHideButton(wwin->frame, WFF_RIGHT_BUTTON);
 
 	wwin->frame->child = wwin;
-	wwin->frame->workspace = w_global.workspace.current;
+	wwin->frame->workspace = wwin->screen_ptr->current_workspace;
 
 #ifdef XKB_BUTTON_HINT
 	if (wPreferences.modelock)
@@ -1629,6 +1660,61 @@ void wWindowUnmap(WWindow *wwin)
 	XUnmapWindow(dpy, wwin->frame->core->window);
 }
 
+void wWindowSingleFocus(WWindow *wwin)
+{
+	WScreen *scr;
+	int x, y, move = 0;
+
+	if (!wwin)
+		return;
+
+	scr = wwin->screen_ptr;
+	wMakeWindowVisible(wwin);
+
+	x = wwin->frame_x;
+	y = wwin->frame_y;
+
+	/* bring window back to visible area */
+	move = wScreenBringInside(scr, &x, &y, wwin->frame->core->width, wwin->frame->core->height);
+
+	if (move) {
+		wWindowConfigure(wwin, x, y, wwin->client.width, wwin->client.height);
+	}
+}
+
+void wWindowFocusPrev(WWindow *wwin, Bool inSameWorkspace)
+{
+	WWindow *tmp;
+
+	if (!wwin || !wwin->prev)
+		return;
+
+	tmp = wwin;
+	while (tmp->prev)
+		tmp = tmp->prev;
+
+	if (inSameWorkspace)
+		while (tmp && (tmp->frame->workspace != wwin->frame->workspace))
+			tmp = tmp->next;
+
+	wWindowSingleFocus(tmp);
+}
+
+void wWindowFocusNext(WWindow *wwin, Bool inSameWorkspace)
+{
+	WWindow *tmp;
+
+	if (!wwin || !wwin->prev)
+		return;
+
+	tmp = wwin->prev;
+	if (inSameWorkspace)
+		while (tmp && (tmp->frame->workspace != wwin->frame->workspace))
+			tmp = tmp->prev;
+
+	wWindowSingleFocus(tmp);
+}
+
 void wWindowFocus(WWindow *wwin, WWindow *owin)
 {
 	WWindow *nowner;
@@ -1866,10 +1952,10 @@ void wWindowChangeWorkspace(WWindow *wwin, int workspace)
 	WApplication *wapp;
 	int unmap = 0;
 
-	if (workspace >= w_global.workspace.count || workspace < 0 || workspace == wwin->frame->workspace)
+	if (workspace >= scr->workspace_count || workspace < 0 || workspace == wwin->frame->workspace)
 		return;
 
-	if (workspace != w_global.workspace.current) {
+	if (workspace != scr->current_workspace) {
 		/* Sent to other workspace. Unmap window */
 		if ((wwin->flags.mapped
 		     || wwin->flags.shaded || (wwin->flags.miniaturized && !wPreferences.sticky_icons))
@@ -1913,23 +1999,23 @@ void wWindowChangeWorkspace(WWindow *wwin, int workspace)
 void wWindowChangeWorkspaceRelative(WWindow *wwin, int amount)
 {
 	WScreen *scr = wwin->screen_ptr;
-	int w = w_global.workspace.current + amount;
+	int w = scr->current_workspace + amount;
 
 	if (amount < 0) {
 		if (w >= 0) {
 			wWindowChangeWorkspace(wwin, w);
 		} else if (wPreferences.ws_cycle) {
-			wWindowChangeWorkspace(wwin, w_global.workspace.count + w);
+			wWindowChangeWorkspace(wwin, scr->workspace_count + w);
 		}
 	} else if (amount > 0) {
-		if (w < w_global.workspace.count) {
+		if (w < scr->workspace_count) {
 			wWindowChangeWorkspace(wwin, w);
 		} else if (wPreferences.ws_advance) {
 			int workspace = WMIN(w, MAX_WORKSPACES - 1);
 			wWorkspaceMake(scr, workspace);
 			wWindowChangeWorkspace(wwin, workspace);
 		} else if (wPreferences.ws_cycle) {
-			wWindowChangeWorkspace(wwin, w % w_global.workspace.count);
+			wWindowChangeWorkspace(wwin, w % scr->workspace_count);
 		}
 	}
 }
@@ -2296,8 +2382,8 @@ void wWindowSaveState(WWindow *wwin)
 	}
 
 	for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
-		if (w_global.shortcut.windows[i] &&
-		    WMCountInArray(w_global.shortcut.windows[i], wwin))
+		if (wwin->screen_ptr->shortcutWindows[i] &&
+		    WMCountInArray(wwin->screen_ptr->shortcutWindows[i], wwin))
 			data[9] |= 1 << i;
 	}
 
@@ -2486,7 +2572,7 @@ void wWindowResetMouseGrabs(WWindow * wwin)
 	/* Mouse grabs can't be done on the client window because of
 	 * ICCCM and because clients that try to do the same will crash.
 	 *
-	 * But there is a problem wich makes tbar buttons of unfocused
+	 * But there is a problem which makes tbar buttons of unfocused
 	 * windows not usable as the click goes to the frame window instead
 	 * of the button itself. Must figure a way to fix that.
 	 */
@@ -2565,7 +2651,19 @@ WMagicNumber wWindowAddSavedState(const char *instance, const char *class,
 	return wstate;
 }
 
-#define SAME(x, y) (((x) && (y) && !strcmp((x), (y))) || (!(x) && !(y)))
+static inline int is_same(const char *x, const char *y)
+{
+	if ((x == NULL) && (y == NULL))
+		return 1;
+
+	if ((x == NULL) || (y == NULL))
+		return 0;
+
+	if (strcmp(x, y) == 0)
+		return 1;
+	else
+		return 0;
+}
 
 WMagicNumber wWindowGetSavedState(Window win)
 {
@@ -2581,8 +2679,9 @@ WMagicNumber wWindowGetSavedState(Window win)
 
 	if (PropGetWMClass(win, &class, &instance)) {
 		while (wstate) {
-			if (SAME(instance, wstate->instance) &&
-			    SAME(class, wstate->class) && SAME(command, wstate->command)) {
+			if (is_same(instance, wstate->instance) &&
+			    is_same(class, wstate->class) &&
+			    is_same(command, wstate->command)) {
 				break;
 			}
 			wstate = wstate->next;
@@ -2754,7 +2853,7 @@ static void titlebarDblClick(WCoreWindow *sender, void *data, XEvent *event)
 				int ndir = dir ^ wwin->flags.maximized;
 
 				if (ndir != 0)
-					wMaximizeWindow(wwin, ndir);
+					wMaximizeWindow(wwin, ndir, wGetHeadForWindow(wwin));
 				else
 					wUnmaximizeWindow(wwin);
 			}
